#include <sys/fcntl.h>
#include <sys/param.h>
#include <sys/cdefs.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#if __FreeBSD_version >= 502100
#include <dev/bktr/ioctl_meteor.h>
#include <dev/bktr/ioctl_bt848.h>
#else
#include <machine/ioctl_meteor.h>
#include <machine/ioctl_bt848.h>
#endif
#include <SDL.h>

// BUILD: cc -Wall -o btviewer -fPIC `sdl11-config --libs --cflags` btviewer.c

const char *usage = "Usage: btviewer -n|s|p -w N -h M\n -n    NTSC\n -s    SECAM\n -p    PAL\n -w    Width\n -h    Height\n";

extern int errno;

caddr_t buf;
int paused = 0;
int sleeptime = 0;
int rows, cols, size, i, c, w, h;

struct meteor_capframe cap;
struct meteor_geomet geo;

SDL_Surface *screen;
SDL_Overlay *overlay;
SDL_Event event;
SDL_Rect rect = { 0 };

void frame_handler() {
    SDL_LockYUVOverlay(overlay);
    memcpy(overlay->pixels[0], buf, size);
    SDL_UnlockYUVOverlay(overlay);    
}

void cleanup() {
    c = METEOR_SIG_MODE_MASK;
    ioctl(i, METEORSSIGNAL, &c);
    c = METEOR_CAP_STOP_CONT;
    ioctl(i, METEORCAPTUR, &c);
    signal(SIGUSR1, SIG_DFL);
    SDL_Quit();
    munmap(buf, size);
    close(i);
}

int main(int argc, char **argv) {
    c = w = h = -1;
    if (argc != 6) {
        printf(usage);
        exit(0);
    }
    
    while ((i = getopt(argc, argv, "nspw:h:")) != -1) {
        if (i == 'n') c = BT848_IFORM_F_NTSCM;
        else if (i == 's') c = BT848_IFORM_F_SECAM;
        else if (i == 'p') c = BT848_IFORM_F_PALBDGHI;
        else if (i == 'w') w = (int) strtol(optarg, NULL, 0);
        else if (i == 'h') h = (int) strtol(optarg, NULL, 0);
        else {
            printf(usage);
            exit(0);
        }
    }
    
    if (w <= 0 || h <= 0) {
        printf(usage);
        exit(0);
    }

    if ((i = open("/dev/bktr", O_RDONLY)) < 0) {
        printf("open bktr failed: %d\n", errno);
        exit(1);
    }

    if (c == BT848_IFORM_F_NTSCM) {
        if (w > 640) cols = 640;
        else cols = w;
        if (h > 480) rows = 480;
        else rows = h;
        // More than 30 fps would be just wasting CPU
        sleeptime = 33333;
    }
    else {
        if (w > 768) cols = 768;
        else cols = w;
        if (h > 576) rows = 576;
        else rows = h;
        sleeptime = 40000;
    }

    size = rows * cols * 2;

    // Capture geometry, we capture full frames (scaling is done by SDL)
    geo.rows = rows;
    geo.columns = cols;
    geo.frames = 1;
    geo.oformat = METEOR_GEO_YUV_PACKED;
    if (ioctl(i, METEORSETGEO, &geo) < 0) {
        printf("ioctl METEORSETGEO failed: %d\n", errno);
        close(i);
        exit(1);
    }

    // Note omission of setting input device (defaults to DEV0)
    // This is considered a tuner function, see bt848 module

    // Set video standard
    if (ioctl(i, BT848SFMT, &c) < 0) {
        printf("ioctl BT848SFMT failed: %d\n", errno);
        close(i);
        exit(1);
    }
    
    // Framebuffer
    buf = mmap((caddr_t)0, size, PROT_READ, MAP_SHARED, i, (off_t)0);
    if (buf == (caddr_t)MAP_FAILED) {
        printf("mmap failed");
        close(i);
        exit (1);
    }
    
    // SDL init and such
    rect.w = w;
    rect.h = h;
    
    if (SDL_Init(SDL_INIT_VIDEO) < 0) {
        printf("SDL_INIT_VIDEO failed: %s", SDL_GetError());
        SDL_Quit();
        munmap(buf, size);
        close(i);
        exit(1);
    }

    if ((screen = SDL_SetVideoMode(rect.w, rect.h, 0, SDL_RESIZABLE)) == NULL) {
        printf("SDL_SetVideoMode failed: %s", SDL_GetError());
        SDL_Quit();
        munmap(buf, size);
        close(i);
        exit(1);
    }
    
    if ((overlay = SDL_CreateYUVOverlay(cols, rows, SDL_UYVY_OVERLAY, screen)) == NULL) {
        printf("SDL_CreateYUVOverlay failed: %s", SDL_GetError());
        SDL_Quit();
        munmap(buf, size);
        close(i);
        exit(1);
    }

    SDL_WM_SetCaption("btviewer", "");
    SDL_DisplayYUVOverlay(overlay, &rect);

    // Set signal handlers
    signal(SIGUSR1, frame_handler);

    // Capture on
    c = METEOR_CAP_CONTINOUS;
    if (ioctl(i, METEORCAPTUR, &c) < 0) {
        printf("ioctl METEORCAPTUR failed: %d\n", errno);        
        signal(SIGUSR1, SIG_DFL);
        SDL_Quit();
        munmap(buf, size);
        close(i);
        exit(1);
    }

    // Set the signal ("got frame")
    c = SIGUSR1;
    if (ioctl(i, METEORSSIGNAL, &c) < 0) {
        printf("ioctl METEORSSIGNAL failed: %d\n", errno);
        c = METEOR_CAP_STOP_CONT;
        ioctl(i, METEORCAPTUR, &c);
        signal(SIGUSR1, SIG_DFL);
        SDL_Quit();
        munmap(buf, size);
        close(i);
        exit(1);
    }
    
    // Clean up at exit
    atexit(cleanup);
    
    // SDL loop starts here
    while (1) {
        if (!paused) SDL_DisplayYUVOverlay(overlay, &rect);
        SDL_PumpEvents();
        // There's three events to handle: quit, pause, resize
        if (SDL_PollEvent(&event)) {
            if (event.type == SDL_KEYDOWN) {
                if (event.key.keysym.sym == SDLK_SPACE) {
                    if (paused) paused = 0;
                    else paused = 1;
                }
                if (event.key.keysym.sym == SDLK_q) break;
            }
            if (event.type == SDL_QUIT) break;
            if (event.type == SDL_VIDEORESIZE) {
                rect.w = event.resize.w;
                rect.h = event.resize.h;
                SDL_SetVideoMode(rect.w, rect.h, 0, SDL_RESIZABLE);
            }
        }
        // Save Our Cycles
        usleep(sleeptime);
    }
    exit(0);
}
