"""
Classes for rendering backends, currently MPlayer and Tvv, builtin for pwc
"""

## MODULE IMPORTS ########################################################

import commands, os, signal
import btcopyright, bthardware, buildprefs

if buildprefs.WITH_PWC:
    import threading, time, pygame, pwc

## MODULE COPYRIGHT ######################################################

MODULE_AUTHOR       = btcopyright.MY_NAME
MODULE_AUTHOR_EMAIL = btcopyright.MY_EMAIL
MODULE_COPYRIGHT    = btcopyright.MY_COPYRIGHT
MODULE_LICENSE      = btcopyright.BSD_LICENSE
MODULE_LICENSE_TEXT = btcopyright.BSD_LICENSE_TEXT

## MODULE CONSTANTS ######################################################

# Commands
BTVIEWER  = "btviewer"
SAAVIEWER = "saaviewer"

# Misc
BT_TUNERDEV  = "/dev/tuner"
SDL_WINDOWID = "SDL_WINDOWID"

## SAAVIEWER CLASS #######################################################

class SaaViewer:
    def __init__(self, wid=-1):
        """
        -> SaaViewer
        
        """
        self.pid = os.getpid()
        self.backend_pid = -1
        self.wid = wid
        if self.wid > 0: 
            os.environ[SDL_WINDOWID] = str(self.wid)
        self.command = SAAVIEWER
        self.args = [self.command, "-w", "400", "-h", "300"]

    def setSize(self, size):
        """
        -> void
        
        Sets the size to use.
        """
        if type(size) == type((1,2)):
            self.args[2] = str(size[0])
            self.args[4] = str(size[1])

    def setWid(self, wid):
        """
        -> void
        
        Sets the window id to wid, for embedding.
        """
        self.wid = wid
        if self.wid > 0: 
            os.environ[SDL_WINDOWID] = str(self.wid)
        else:
            if SDL_WINDOWID in os.environ.keys():
                del os.environ[SDL_WINDOWID]
        
    def pid(self):
        """
        -> int

        Returns the PID of this process (the backend's parent).
        """
        return self.pid
	    
    def backendPID(self):
        """
        -> int

        Returns the PID of the backend process, -1 if none is running.
        """
        return self.backend_pid

    def start(self):
        """
        -> bool

        Attempts to start the backend and get its PID. Returns True on
        success, False on failure (presumably it's not running).
        """
        try:
            self.backend_pid = os.spawnvp(os.P_NOWAIT, self.command, self.args)
            return True
        except OSError:
            # Backend never started
            self.backend_pid = -1
            return False 

    def stop(self):
        """
        -> bool

        Attempts to stop the backend and set its PID to -1. Returns True
        on success, False on failure (presumably it's still running).
        """
        if self.backend_pid != -1:
            try:
                os.kill(self.backend_pid, signal.SIGTERM)
                self.backend_pid = -1
                return True
            except OSError, IOError:
                return False
        else:
            return True

## BTVIEWER CLASS ########################################################

class BtViewer(SaaViewer):
    """
    """
    def __init__(self, wid=-1):
        """
        -> BtViewer
        
        """
        SaaViewer.__init__(self, wid)
        self.command = BTVIEWER
        self.args = [self.command, "-w", "400", "-h", "300"]
        norm = bthardware.bktr_norm()
        if norm == bthardware.NORM_NTSC: self.args.append("-n")
        elif norm == bthardware.NORM_SECAM: self.args.append("-s")
        else: self.args.append("-p")
        self.fd = None

    def start(self):
        res = SaaViewer.start(self)
        if res:
            self.fd = open(BT_TUNERDEV, "r")
        return res

    def stop(self):
        res = SaaViewer.stop(self)
        if res:
            self.fd.close()
        return res

## PWCPLAYER CLASS #######################################################

class PWCPlayer:
    """
    This class describes a renderer for webcams supported by the pwc driver.
    It is "native" to kbtv and uses the pygame framework (which is SDL based).
    The preferred capture size and framerate can be set with the pwc module.
    This player inquires the pwc driver on init about the size, but not the
    framerate (doesn't seem nescessary as updating the overlay can easily
    be done fast enough to provide full 25 or 30 frames per second). In
    practice a framerate of 15 with a 320x240 sized image works fine.
    
    This renderer is still a bit crude. It is to serve as a first step towards
    a unified SDL based renderer for all drivers.

    TODO: Move to C code.

    """
    def __init__(self, fd, size, wid=-1):
        """
        The constructor expects a file descriptor to the already opened
        device (/dev/video0). It will also accept a window ID as the second
        argument, and if not -1, the rendering pygame window embeds itself 
        into the widget with that wid.Size refers to the display size, which
        need not be the same as the capture size.
        """
        self.playing = False
        self.grab = False
        self.fd = fd
        self.wid = wid
        s = pwc.camera_capsize(self.fd.fileno())
        if s == 0: self.capsize = (128, 96)
        elif s == 1: self.capsize = (160, 120)
        elif s == 2: self.capsize = (176, 144)
        elif s == 3: self.capsize = (320, 240)
        elif s == 4: self.capsize = (352, 288)
        elif s == 5: self.capsize = (640, 480)
        else: 
            self.capsize = (320, 240)
        self.size = size # size is displaysize
        # For every Y plane of w*h size there is a U and a V plane of w*h/4
        # So that makes image data size of 6/4 * (320, 240) 
        self.datasize = self.capsize[0] * self.capsize[1] * 3 / 2
        # 4:1:1 planar, actually YVU not YUV (YYYYVU)
        # dataX = end of plane
        self.dataY = self.capsize[0] * self.capsize[1]
        self.dataV = self.capsize[0] * self.capsize[1] * 5 / 4
        if wid != -1:
            # Embedding
            os.environ['SDL_WINDOWID'] = str(self.wid)
        pygame.display.init()
        self.display = pygame.display.set_mode(self.size, pygame.RESIZABLE)
        self.overlay = pygame.overlay.Overlay(pygame.IYUV_OVERLAY, self.capsize)

    def setSize(self, size):
        """
        -> void
        
        Set display size. Size is a tuple (w,h).
        """
        self.size = size
        self.pause()
        self.display = pygame.display.set_mode(self.size, pygame.RESIZABLE)
        self.overlay = pygame.overlay.Overlay(pygame.IYUV_OVERLAY, self.capsize)
        self.play()

    def readImageData(self):
        """
        -> (y, u, v)
        
        Reads image from device, returns its data orded in a y, u, v tuple,
        or None if reading failed.
        """
        data = os.read(self.fd.fileno(), self.datasize)
        if data:
            (y, v, u) = (data[:self.dataY], data[self.dataY:self.dataV],
                         data[self.dataV:])
            return (y, u, v)
        return None

    def displayImage(self):
        """
        -> void
        
        Reads image data, displays it on the overlay, and if the
        attribute playing is still True it triggers a timer that calls
        this method again for the next read. It returns the yuv data 
        tuple (or None), so it may be processed in some way.
        """
        yuv = self.readImageData()
        if type(yuv) == type((1,2,3)):
            self.overlay.set_location((0, 0), self.size)
            self.overlay.display((yuv[0], yuv[1], yuv[2]))
            if self.playing:
                self.update()

    def play(self):
        """
        -> void
        
        Starts the renderer.
        """
        self.playing = True
        self.update()

    def update(self):
        """
        -> void
        
        Starts a timer to run the displayImage() method. For now, it does
        so at fixed 30 ms intervals.
        """
        t = threading.Timer(0.03, self.displayImage)
        t.setDaemon(True)
        t.start()

    def pause(self):
        """
        -> void
        
        Pauses the renderer.
        """
        self.playing = False
        time.sleep(0.1)
    
    def stop(self):
        self.pause()
        pygame.quit()

## END ###################################################################

if __name__ == "__main__":
    print "This is a module. Import it instead."
