#include <sys/fcntl.h>
#include <sys/param.h>
#include <sys/cdefs.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include "saa/kmod/saa713x_ioctl.h"
#include <SDL.h>

// BUILD: cc [-Wall] -o saaviewer -fPIC `sdl11-config --libs --cflags` saaviewer.c

const char *usage = "Usage: saaviewer -w N -h M\n -w    Width\n -h    Height\n";

extern int errno;

caddr_t buf;
int paused = 0;
int sleeptime = 0;
int rows, cols, size, i, c, w, h, bpp, empty;

struct saa_video_std std;
struct saa_video_spec spec;

SDL_Surface *screen;
SDL_Overlay *overlay;
SDL_Event event;
SDL_Rect rect = { 0 };

void frame_handler() {
    SDL_LockYUVOverlay(overlay);
    memcpy(overlay->pixels[0], buf, size);
    SDL_UnlockYUVOverlay(overlay);
}

void cleanup() {
    ioctl(i, CAPTUREOFF, &empty);
    signal(SIGUSR1, SIG_DFL);
    SDL_Quit();
    munmap(buf, size);
    close(i);
}

int main(int argc, char **argv) {
    w = h = -1;
    if (argc != 5) {
        printf(usage);
        exit(0);
    }
    
    while ((i = getopt(argc, argv, "nspw:h:")) != -1) {
        if (i == 'w') w = (int) strtol(optarg, NULL, 0);
        else if (i == 'h') h = (int) strtol(optarg, NULL, 0);
        else {
            printf(usage);
            exit(0);
        }
    }

    if (w <= 0 || h <= 0) {
        printf(usage);
        exit(0);
    }
    
    if ((i = open("/dev/saa0", O_RDWR)) < 0) {
        printf("open saa failed: %d\n", errno);
        exit(1);
    }
    
    if (ioctl(i, GETVIDEOSTD, &std) < 0) {
        printf("ioctl GETVIDEOSTD failed: %d\n", errno);
        close(i);
        exit(1);
    }
    
    // SDTV PAL B/G/I..: 720x576i NTSCM/PALM: 720x480i
    // Capture at full frame size, regardless of display size
    rows = (std.std_video_yinstop - std.std_video_yinstart + 1) * 2;
    cols = std.std_video_xinstop - std.std_video_xinstart + 1;
    size = rows * cols * 2;
    if (rows == 480) sleeptime = 33333;
    else sleeptime = 40000;

    // Video spec
    spec.svs_cap_nrtasks = 1;
    spec.svs_vfmt = VFMT_YUV422;
    spec.svs_cap_swapbytes = SWAPBYTES_NONE;
    spec.svs_flags = VSPEC_FLAG_CAPTURE_SIGNAL;
    spec.svs_cap_bufsz = SAA_VCAP_FRAME_MAX_SZ;
    spec.svs_buf_base = 0;
    spec.svs_buf_width = cols;
    spec.svs_buf_height = rows * 2;
    spec.svs_buf_offset[0][0] = 0;
    bpp = (spec.svs_vfmt & VFMT_BPP_MASK) >> VFMT_BPP_SHIFT; // not needed
    spec.svs_buf_pitch[0] = spec.svs_buf_width * bpp / 8;

    // SDL init and such
    rect.w = w;
    rect.h = h;

    if (SDL_Init(SDL_INIT_VIDEO) < 0) {
        printf("SDL_INIT_VIDEO failed: %s", SDL_GetError());
        SDL_Quit();
        close(i);
        exit(1);
    }

    if ((screen = SDL_SetVideoMode(rect.w, rect.h, 0, SDL_RESIZABLE)) == NULL) {
        printf("SDL_SetVideoMode failed: %s", SDL_GetError());
        SDL_Quit();
        close(i);
        exit(1);
    }
    
    if ((overlay = SDL_CreateYUVOverlay(cols, rows, SDL_UYVY_OVERLAY, screen)) == NULL) {
        printf("SDL_CreateYUVOverlay failed: %s", SDL_GetError());
        SDL_Quit();
        close(i);
        exit(1);
    }

    SDL_WM_SetCaption("saaviewer", "");
    SDL_DisplayYUVOverlay(overlay, &rect);

    // Set signal handler
    signal(SIGUSR1, frame_handler);

    // Capture on
    if (ioctl(i, CAPTUREON, &spec) < 0) {
        printf("ioctl CAPTUREON failed: %d\n", errno);
        close(i);
        exit(1);
    }

    // Framebuffer
    //buf = mmap(0, SAA_VCAP_FRAME_MAX_SZ, PROT_READ, 0, i, SAA_MMAP_T0_OFFSET);
    buf = mmap((caddr_t)0, size, PROT_READ, 0, i, (off_t)SAA_MMAP_T0_OFFSET);
    if (buf == (caddr_t)MAP_FAILED) {
        printf("mmap failed");
        close(i);
        exit (1);
    }

    // Cleanup at exit
    atexit(cleanup);

    // SDL loop starts here
    while (1) {
        if (!paused) SDL_DisplayYUVOverlay(overlay, &rect);
        SDL_PumpEvents();
        // There's three events to handle: quit, pause, resize
        if (SDL_PollEvent(&event)) {
            if (event.type == SDL_KEYDOWN) {
                if (event.key.keysym.sym == SDLK_SPACE) {
                    if (paused) paused = 0;
                    else paused = 1;
                }
                if (event.key.keysym.sym == SDLK_q) break;
            }
            if (event.type == SDL_QUIT) break;
            if (event.type == SDL_VIDEORESIZE) {
                rect.w = event.resize.w;
                rect.h = event.resize.h;
                SDL_SetVideoMode(rect.w, rect.h, 0, SDL_RESIZABLE);
            }
        }
        // Save Our Cycles
        usleep(sleeptime);
    }
    exit(0);
}
